# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definition of PipelineChannel."""

import abc
import contextlib
import dataclasses
import json
import re
from typing import Dict, List, Optional, Union

from kfp.components.types import type_utils


@dataclasses.dataclass
class ConditionOperator:
    """Represents a condition expression to be used in dsl.Condition().

    Attributes:
      operator: The operator of the condition.
      left_operand: The left operand.
      right_operand: The right operand.
    """
    operator: str
    left_operand: Union['PipelineParameterChannel', type_utils.PARAMETER_TYPES]
    right_operand: Union['PipelineParameterChannel', type_utils.PARAMETER_TYPES]


# The string template used to generate the placeholder of a PipelineChannel.
_PIPELINE_CHANNEL_PLACEHOLDER_TEMPLATE = (
    '{{channel:task=%s;name=%s;type=%s;}}')
# The regex for parsing PipelineChannel placeholders from a string.
_PIPELINE_CHANNEL_PLACEHOLDER_REGEX = (
    r'{{channel:task=([\w\s_-]*);name=([\w\s_-]+);type=([\w\s{}":_-]*);}}')


class PipelineChannel(abc.ABC):
    """Represents a future value that is passed between pipeline components.

    A PipelineChannel object can be used as a pipeline function argument so that
    it will be a pipeline artifact or parameter that shows up in ML Pipelines
    system UI. It can also represent an intermediate value passed between
    components.

    Attributes:
        name: The name of the pipeline channel.
        channel_type: The type of the pipeline channel.
        task_name: The name of the task that produces the pipeline channel.
            None means it is not produced by any task, so if None, either user
            constructs it directly (for providing an immediate value), or it is
            a pipeline function argument.
        pattern: The serialized string regex pattern this pipeline channel
            created from.
    """

    @abc.abstractmethod
    def __init__(
        self,
        name: str,
        channel_type: Union[str, Dict],
        task_name: Optional[str] = None,
    ):
        """Initializes a PipelineChannel instance.

        Args:
            name: The name of the pipeline channel. The name will be sanitized
                to be k8s compatible.
            channel_type: The type of the pipeline channel.
            task_name: Optional; The name of the task that produces the pipeline
                channel. If provided, the task name will be sanitized to be k8s
                compatible.

        Raises:
            ValueError: If name or task_name contains invalid characters.
            ValueError: If both task_name and value are set.
        """
        valid_name_regex = r'^[A-Za-z][A-Za-z0-9\s_-]*$'
        if not re.match(valid_name_regex, name):
            raise ValueError(
                f'Only letters, numbers, spaces, "_", and "-" are allowed in the name. Must begin with a letter. Got name: {name}'
            )

        self.name = name
        self.channel_type = channel_type
        # ensure value is None even if empty string or empty list/dict
        # so that serialization and unserialization remain consistent
        # (i.e. None => '' => None)
        self.task_name = task_name or None
        from kfp.components import pipeline_context

        default_pipeline = pipeline_context.Pipeline.get_default_pipeline()
        if self.task_name is not None and default_pipeline is not None and default_pipeline.tasks:
            self.task = pipeline_context.Pipeline.get_default_pipeline().tasks[
                self.task_name]
        else:
            self.task = None

    @property
    def full_name(self) -> str:
        """Unique name for the PipelineChannel."""
        return f'{self.task_name}-{self.name}' if self.task_name else self.name

    @property
    def pattern(self) -> str:
        """Unique pattern for the PipelineChannel."""
        return str(self)

    def __str__(self) -> str:
        """String representation of the PipelineChannel.

        The string representation is a string identifier so we can mix
        the PipelineChannel inline with other strings such as arguments.
        For example, we can support: ['echo %s' % param] as the
        container command and later a compiler can replace the
        placeholder '{{pipeline_channel:task=%s;name=%s;type=%s}}' with
        its own parameter identifier.
        """
        task_name = self.task_name or ''
        name = self.name
        channel_type = self.channel_type or ''
        if isinstance(channel_type, dict):
            channel_type = json.dumps(channel_type)
        return _PIPELINE_CHANNEL_PLACEHOLDER_TEMPLATE % (task_name, name,
                                                         channel_type)

    def __repr__(self) -> str:
        """Representation of the PipelineChannel.

        We make repr return the placeholder string so that if someone
        uses str()-based serialization of complex objects containing
        `PipelineChannel`, it works properly. (e.g. str([1, 2, 3,
        kfp.pipeline_channel.PipelineParameterChannel("aaa"), 4, 5, 6,]))
        """
        return str(self)

    def __hash__(self) -> int:
        """Returns the hash of a PipelineChannel."""
        return hash(self.pattern)

    def __eq__(self, other):
        return ConditionOperator('==', self, other)

    def __ne__(self, other):
        return ConditionOperator('!=', self, other)

    def __lt__(self, other):
        return ConditionOperator('<', self, other)

    def __le__(self, other):
        return ConditionOperator('<=', self, other)

    def __gt__(self, other):
        return ConditionOperator('>', self, other)

    def __ge__(self, other):
        return ConditionOperator('>=', self, other)


class PipelineParameterChannel(PipelineChannel):
    """Represents a pipeline parameter channel.

    Attributes:
      name: The name of the pipeline channel.
      channel_type: The type of the pipeline channel.
      task_name: The name of the task that produces the pipeline channel.
        None means it is not produced by any task, so if None, either user
        constructs it directly (for providing an immediate value), or it is a
        pipeline function argument.
      pattern: The serialized string regex pattern this pipeline channel created
        from.
      value: The actual value of the pipeline channel. If provided, the
        pipeline channel is "resolved" immediately.
    """

    def __init__(
        self,
        name: str,
        channel_type: Union[str, Dict],
        task_name: Optional[str] = None,
        value: Optional[type_utils.PARAMETER_TYPES] = None,
    ):
        """Initializes a PipelineArtifactChannel instance.

        Args:
          name: The name of the pipeline channel.
          channel_type: The type of the pipeline channel.
          task_name: Optional; The name of the task that produces the pipeline
            channel.
          value: Optional; The actual value of the pipeline channel.

        Raises:
          ValueError: If name or task_name contains invalid characters.
          ValueError: If both task_name and value are set.
          TypeError: If the channel type is not a parameter type.
        """
        if task_name and value:
            raise ValueError('task_name and value cannot be both set.')

        if not type_utils.is_parameter_type(channel_type):
            raise TypeError(f'{channel_type} is not a parameter type.')

        self.value = value

        super(PipelineParameterChannel, self).__init__(
            name=name,
            channel_type=channel_type,
            task_name=task_name,
        )


class PipelineArtifactChannel(PipelineChannel):
    """Represents a pipeline artifact channel.

    Attributes:
      name: The name of the pipeline channel.
      channel_type: The type of the pipeline channel.
      task_name: The name of the task that produces the pipeline channel.
        A pipeline artifact channel is always produced by some task.
      pattern: The serialized string regex pattern this pipeline channel created
        from.
    """

    def __init__(
        self,
        name: str,
        channel_type: Union[str, Dict],
        task_name: Optional[str],
        is_artifact_list: bool,
    ):
        """Initializes a PipelineArtifactChannel instance.

        Args:
            name: The name of the pipeline channel.
            channel_type: The type of the pipeline channel.
            task_name: Optional; the name of the task that produces the pipeline
                channel.

        Raises:
            ValueError: If name or task_name contains invalid characters.
            TypeError: If the channel type is not an artifact type.
        """
        if type_utils.is_parameter_type(channel_type):
            raise TypeError(f'{channel_type} is not an artifact type.')

        self.is_artifact_list = is_artifact_list

        super(PipelineArtifactChannel, self).__init__(
            name=name,
            channel_type=channel_type,
            task_name=task_name,
        )


def create_pipeline_channel(
    name: str,
    channel_type: Union[str, Dict],
    task_name: Optional[str] = None,
    value: Optional[type_utils.PARAMETER_TYPES] = None,
    is_artifact_list: bool = False,
) -> PipelineChannel:
    """Creates a PipelineChannel object.

    Args:
        name: The name of the channel.
        channel_type: The type of the channel, which decides whether it is an
            PipelineParameterChannel or PipelineArtifactChannel
        task_name: Optional; the task that produced the channel.
        value: Optional; the realized value for a channel.

    Returns:
        A PipelineParameterChannel or PipelineArtifactChannel object.
    """
    if type_utils.is_parameter_type(channel_type):
        return PipelineParameterChannel(
            name=name,
            channel_type=channel_type,
            task_name=task_name,
            value=value,
        )
    else:
        return PipelineArtifactChannel(
            name=name,
            channel_type=channel_type,
            task_name=task_name,
            is_artifact_list=is_artifact_list,
        )


def extract_pipeline_channels_from_string(
        payload: str) -> List[PipelineChannel]:
    """Extracts a list of PipelineChannel instances from the payload string.

    Note: this function removes all duplicate matches.

    Args:
      payload: A string that may contain serialized PipelineChannels.

    Returns:
      A list of PipelineChannels found from the payload.
    """
    matches = re.findall(_PIPELINE_CHANNEL_PLACEHOLDER_REGEX, payload)
    unique_channels = set()
    for match in matches:
        task_name, name, channel_type = match

        # channel_type could be either a string (e.g. "Integer") or a dictionary
        # (e.g.: {"custom_type": {"custom_property": "some_value"}}).
        # Try loading it into dictionary, if failed, it means channel_type is a
        # string.
        with contextlib.suppress(json.JSONDecodeError):
            channel_type = json.loads(channel_type)

        if type_utils.is_parameter_type(channel_type):
            pipeline_channel = PipelineParameterChannel(
                name=name,
                channel_type=channel_type,
                task_name=task_name,
            )
        else:
            pipeline_channel = PipelineArtifactChannel(
                name=name,
                channel_type=channel_type,
                task_name=task_name,
                # currently no support for getting the index from a list of artifacts (e.g., my_datasets[0].uri), so this will always be False until accessing a single artifact element is supported
                is_artifact_list=False,
            )
        unique_channels.add(pipeline_channel)

    return list(unique_channels)


def extract_pipeline_channels_from_any(
    payload: Union[PipelineChannel, str, list, tuple, dict]
) -> List[PipelineChannel]:
    """Recursively extract PipelineChannels from any object or list of objects.

    Args:
      payload: An object that contains serialized PipelineChannels or k8
        definition objects.

    Returns:
      A list of PipelineChannels found from the payload.
    """
    if not payload:
        return []

    if isinstance(payload, PipelineChannel):
        return [payload]

    if isinstance(payload, str):
        return list(set(extract_pipeline_channels_from_string(payload)))

    if isinstance(payload, (list, tuple)):
        pipeline_channels = []
        for item in payload:
            pipeline_channels += extract_pipeline_channels_from_any(item)
        return list(set(pipeline_channels))

    if isinstance(payload, dict):
        pipeline_channels = []
        for key, value in payload.items():
            pipeline_channels += extract_pipeline_channels_from_any(key)
            pipeline_channels += extract_pipeline_channels_from_any(value)
        return list(set(pipeline_channels))

    # TODO(chensun): extract PipelineChannel from v2 container spec?

    return []



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definition for PipelineTaskFinalStatus."""

import dataclasses
from typing import Optional


@dataclasses.dataclass
class PipelineTaskFinalStatus:
    """A final status of a pipeline task. Annotate a component parameter with
    this class to obtain a handle to a task's status (see example).

    This is the Python representation of the proto message `PipelineTaskFinalStatus <https://github.com/kubeflow/pipelines/blob/d8b9439ef92b88da3420df9e8c67db0f1e89d4ef/api/v2alpha1/pipeline_spec.proto#L929-L951>`_.

    Examples:
      ::

        @dsl.component
        def task_status(user_input: str, status: PipelineTaskFinalStatus):
            print('Pipeline status: ', status.state)
            print('Job resource name: ', status.pipeline_job_resource_name)
            print('Pipeline task name: ', status.pipeline_task_name)
            print('Error code: ', status.error_code)
            print('Error message: ', status.error_message)

        @dsl.pipeline(name='my_pipeline')
        def my_pipeline():
            task = task_status(user_input='my_input')
    """
    state: str
    """Final state of the task. The value could be one of ``'SUCCEEDED'``, ``'FAILED'`` or ``'CANCELLED'``."""

    pipeline_job_resource_name: str
    """Pipeline job resource name, in the format of ``projects/{project}/locations/{location}/pipelineJobs/{pipeline_job}``."""

    pipeline_task_name: str
    """Name of the task that produced this status."""

    error_code: Optional[int]
    """The `google.rpc.Code <github.com/googleapis/googleapis/blob/master/google/rpc/code.proto>`_ in case of error. If state is ``'SUCCEEDED'``, this is ``None``."""

    error_message: Optional[str]
    """In case of error, the detailed error message. If state is ``'SUCCEEDED'``, this is ``None``."""



# Copyright 2020-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import datetime
import enum
import json
from typing import Any, Dict

import click
import kfp_server_api
import tabulate

KFP_TABLE_FORMAT = 'custom-simple'

tabulate._table_formats.update({  # type: ignore
    KFP_TABLE_FORMAT:
        tabulate.TableFormat(
            lineabove=None,
            linebelowheader=None,
            linebetweenrows=None,
            linebelow=None,
            headerrow=tabulate.DataRow('', '  ', ''),
            datarow=tabulate.DataRow('', '  ', ''),
            padding=0,
            with_header_hide=['lineabove', 'linebelow'])
})


@enum.unique
class OutputFormat(enum.Enum):
    """Enumerated class with the allowed output format constants."""
    table = 'table'
    json = 'json'


def snake_to_header(string: str) -> str:
    """Converts a snake case string to a table header by replacing underscores
    with spaces and making uppercase.

    Args:
        string (str): The snake case string.

    Returns:
        str: The header.
    """
    return string.replace('_', ' ').upper()


@dataclasses.dataclass
class ExperimentData:
    id: str
    name: str
    created_at: str
    storage_state: str


def transform_experiment(
        exp: kfp_server_api.V2beta1Experiment) -> Dict[str, Any]:
    return dataclasses.asdict(
        ExperimentData(
            id=exp.experiment_id,
            name=exp.display_name,
            created_at=exp.created_at.isoformat(),
            storage_state=exp.storage_state))


@dataclasses.dataclass
class PipelineData:
    id: str
    name: str
    created_at: str


def transform_pipeline(
        pipeline: kfp_server_api.V2beta1Pipeline) -> Dict[str, Any]:
    return dataclasses.asdict(
        PipelineData(
            id=pipeline.pipeline_id,
            name=pipeline.display_name,
            created_at=pipeline.created_at.isoformat(),
        ))


@dataclasses.dataclass
class PipelineVersionData:
    id: str
    name: str
    created_at: str
    parent_id: str


def transform_pipeline_version(
        pipeline_version: kfp_server_api.V2beta1PipelineVersion
) -> Dict[str, Any]:
    return dataclasses.asdict(
        PipelineVersionData(
            id=pipeline_version.pipeline_version_id,
            name=pipeline_version.display_name,
            created_at=pipeline_version.created_at.isoformat(),
            parent_id=pipeline_version.pipeline_id,
        ))


@dataclasses.dataclass
class RunData:
    id: str
    name: str
    created_at: str
    state: str
    storage_state: str


def transform_run(run: kfp_server_api.V2beta1Run) -> Dict[str, Any]:
    return dataclasses.asdict((RunData(
        id=run.run_id,
        name=run.display_name,
        created_at=run.created_at.isoformat(),
        state=run.state,
        storage_state=run.storage_state,
    )))


@dataclasses.dataclass
class RecurringRunData:
    id: str
    name: str
    created_at: str
    experiment_id: str
    status: str


def transform_recurring_run(
        recurring_run: kfp_server_api.V2beta1RecurringRun) -> Dict[str, Any]:
    return dataclasses.asdict(
        RecurringRunData(
            id=recurring_run.recurring_run_id,
            name=recurring_run.display_name,
            created_at=recurring_run.created_at.isoformat(),
            experiment_id=recurring_run.experiment_id,
            status=recurring_run.status))


@enum.unique
class ModelType(enum.Enum):
    """Enumerated class with the allowed output format constants."""
    EXPERIMENT = 'EXPERIMENT'
    PIPELINE = 'PIPELINE'
    PIPELINE_VERSION = 'PIPELINE_VERSION'
    RUN = 'RUN'
    RECURRING_RUN = 'RECURRING_RUN'


transformer_map = {
    ModelType.EXPERIMENT: transform_experiment,
    ModelType.PIPELINE: transform_pipeline,
    ModelType.PIPELINE_VERSION: transform_pipeline_version,
    ModelType.RUN: transform_run,
    ModelType.RECURRING_RUN: transform_recurring_run,
}

dataclass_map = {
    ModelType.EXPERIMENT: ExperimentData,
    ModelType.PIPELINE: PipelineData,
    ModelType.PIPELINE_VERSION: PipelineVersionData,
    ModelType.RUN: RunData,
    ModelType.RECURRING_RUN: RecurringRunData,
}


class DatetimeEncoder(json.JSONEncoder):
    """JSON encoder for serializing datetime objects."""

    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def print_output(resources: list, model_type: ModelType,
                 output_format: str) -> None:
    """Prints output in tabular or JSON format, using click.echo.

    Args:
        resources (list): List of same-type resources to print.
        output_format (str): One of 'table' or 'json'.

    Raises:
        NotImplementedError: If the output format is not one of 'table' or 'json'.
    """
    if isinstance(resources, list):
        single_resource = False
    else:
        resources = [resources]
        single_resource = True

    if output_format == OutputFormat.table.name:
        transformer = transformer_map[model_type]
        output_headers = dataclass_map[  # type: ignore
            model_type].__dataclass_fields__.keys()
        resources = [transformer(r) for r in resources]

        data = [list(resource.values()) for resource in resources]
        headers = [snake_to_header(header) for header in output_headers]
        click.echo(
            tabulate.tabulate(data, headers=headers, tablefmt='custom-simple'))

    elif output_format == OutputFormat.json.name:
        data = resources[0].to_dict() if single_resource else [
            resources.to_dict() for resources in resources
        ]
        click.echo(json.dumps(data, indent=2, cls=DatetimeEncoder), nl=False)
    else:
        raise NotImplementedError(f'Unknown output format: {output_format}.')


def print_deleted_text(resource_type: str, resource_id: str,
                       output_format: str) -> None:
    """Prints a standardized output for deletion actions, using click.echo.

    Args:
        resource_type (str): The type of resource (e.g. 'experiment') deleted.
        resource_id (str): The ID of the resource deleted.
        output_format (str): The format for the output (one of 'table' or 'json').

    Raises:
        NotImplementedError: If the output format is not one of 'table' or 'json'.
    """
    if output_format == OutputFormat.table.name:
        click.echo(f'{resource_type.capitalize()} {resource_id} deleted.')

    elif output_format == OutputFormat.json.name:
        click.echo(json.dumps(resource_id, indent=2), nl=False)

    else:
        raise NotImplementedError(f'Unknown output format: {output_format}.')



# Copyright 2019 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    'ContainerBuilder',
]

import logging
import tarfile
import tempfile
import os
import uuid

SERVICEACCOUNT_NAMESPACE = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
GCS_STAGING_BLOB_DEFAULT_PREFIX = 'kfp_container_build_staging'
GCR_DEFAULT_IMAGE_SUFFIX = 'kfp_container'
KANIKO_EXECUTOR_IMAGE_DEFAULT = 'gcr.io/kaniko-project/executor@sha256:78d44ec4e9cb5545d7f85c1924695c89503ded86a59f92c7ae658afa3cff5400'


def _get_project_id():
    import requests
    URL = "http://metadata.google.internal/computeMetadata/v1/project/project-id"
    headers = {'Metadata-Flavor': 'Google'}
    r = requests.get(url=URL, headers=headers)
    if not r.ok:
        raise RuntimeError(
            'ContainerBuilder failed to retrieve the project id.')
    return r.text


def _get_instance_id():
    import requests
    URL = "http://metadata.google.internal/computeMetadata/v1/instance/id"
    headers = {'Metadata-Flavor': 'Google'}
    r = requests.get(url=URL, headers=headers)
    if not r.ok:
        raise RuntimeError(
            'ContainerBuilder failed to retrieve the instance id.')
    return r.text


class ContainerBuilder(object):
    """ContainerBuilder helps build a container image."""

    def __init__(self,
                 gcs_staging=None,
                 default_image_name=None,
                 namespace=None,
                 service_account='kubeflow-pipelines-container-builder',
                 kaniko_executor_image=KANIKO_EXECUTOR_IMAGE_DEFAULT,
                 k8s_client_configuration=None):
        """
    Args:
      gcs_staging (str): GCS bucket/blob that can store temporary build files,
          default is gs://PROJECT_ID/kfp_container_build_staging. You have to
          specify this when it doesn't run in cluster.
      default_image_name (str): Target container image name that will be used by the build method if the target_image argument is not specified.
      namespace (str): Kubernetes namespace where the container builder pod is launched,
          default is the same namespace as the notebook service account in cluster
          or 'kubeflow' if not in cluster. If using the full Kubeflow
          deployment and not in cluster, you should specify your own user namespace.
      service_account (str): Kubernetes service account the pod uses for container building,
          The default value is "kubeflow-pipelines-container-builder". It works with Kubeflow Pipelines clusters installed using Google Cloud Marketplace or Standalone with version > 0.4.0.
          The service account should have permission to read and write from staging gcs path and upload built images to gcr.io.
      kaniko_executor_image (str): Docker image used to run kaniko executor. Defaults to gcr.io/kaniko-project/executor:v0.10.0.
      k8s_client_configuration (kubernetes.Configuration): Kubernetes client configuration object to be used when talking with Kubernetes API.
        This is optional. If not specified, it will use the default configuration. This can be used to personalize the client used to talk to the Kubernetes server and change authentication parameters.
    """
        self._gcs_staging = gcs_staging
        self._gcs_staging_checked = False
        self._default_image_name = default_image_name
        self._namespace = namespace
        self._service_account = service_account
        self._kaniko_image = kaniko_executor_image
        self._k8s_client_configuration = k8s_client_configuration

    def _get_namespace(self):
        if self._namespace is None:
            # Configure the namespace
            if os.path.exists(SERVICEACCOUNT_NAMESPACE):
                with open(SERVICEACCOUNT_NAMESPACE, 'r') as f:
                    self._namespace = f.read()
            else:
                self._namespace = 'kubeflow'
        return self._namespace

    def _get_staging_location(self):
        if self._gcs_staging_checked:
            return self._gcs_staging

        # Configure the GCS staging bucket
        if self._gcs_staging is None:
            try:
                gcs_bucket = _get_project_id()
            except:
                raise ValueError(
                    'Cannot get the Google Cloud project ID, please specify the gcs_staging argument.'
                )
            self._gcs_staging = 'gs://' + gcs_bucket + '/' + GCS_STAGING_BLOB_DEFAULT_PREFIX
        else:
            from pathlib import PurePath
            path = PurePath(self._gcs_staging).parts
            if len(path) < 2 or not path[0].startswith('gs'):
                raise ValueError('Error: {} should be a GCS path.'.format(
                    self._gcs_staging))
            gcs_bucket = path[1]
        from ._gcs_helper import GCSHelper
        GCSHelper.create_gcs_bucket_if_not_exist(gcs_bucket)
        self._gcs_staging_checked = True
        return self._gcs_staging

    def _get_default_image_name(self):
        if self._default_image_name is None:
            # KubeFlow Jupyter notebooks have environment variable with the notebook ID
            try:
                nb_id = os.environ.get('NB_PREFIX', _get_instance_id())
            except:
                raise ValueError('Please provide the default_image_name.')
            nb_id = nb_id.replace('/', '-').strip('-')
            self._default_image_name = os.path.join('gcr.io', _get_project_id(),
                                                    nb_id,
                                                    GCR_DEFAULT_IMAGE_SUFFIX)
        return self._default_image_name

    def _generate_kaniko_spec(self, context, docker_filename, target_image):
        """_generate_kaniko_yaml generates kaniko job yaml based on a template
        yaml."""
        content = {
            'apiVersion': 'v1',
            'metadata': {
                'generateName': 'kaniko-',
                'namespace': self._get_namespace(),
                'annotations': {
                    'sidecar.istio.io/inject': 'false'
                },
            },
            'kind': 'Pod',
            'spec': {
                'restartPolicy': 'Never',
                'containers': [{
                    'name': 'kaniko',
                    'args': [
                        '--cache=true',
                        '--dockerfile=' + docker_filename,
                        '--context=' + context,
                        '--destination=' + target_image,
                        '--digest-file=/dev/termination-log',  # This is suggested by the Kaniko devs as a way to return the image digest from Kaniko Pod. See https://github.com/GoogleContainerTools/kaniko#--digest-file
                    ],
                    'image': self._kaniko_image,
                }],
                'serviceAccountName': self._service_account
            }
        }
        return content

    def _wrap_dir_in_tarball(self, tarball_path, dir_name):
        """_wrap_files_in_tarball creates a tarball for all the files in the
        directory."""
        if not tarball_path.endswith('.tar.gz'):
            raise ValueError('the tarball path should end with .tar.gz')
        with tarfile.open(tarball_path, 'w:gz') as tarball:
            tarball.add(dir_name, arcname='')

    def build(self,
              local_dir,
              docker_filename: str = 'Dockerfile',
              target_image=None,
              timeout=1000):
        """
    Args:
      local_dir (str): local directory that stores all the necessary build files
      docker_filename (str): the path of the Dockerfile relative to the local_dir
      target_image (str): The container image name where the data will be pushed. Can include tag. If not specified, the function will use the default_image_name specified when creating ContainerBuilder.
      timeout (int): time out in seconds. Default: 1000
    """
        target_image = target_image or self._get_default_image_name()
        # Prepare build context
        with tempfile.TemporaryDirectory() as local_build_dir:
            from ._gcs_helper import GCSHelper
            logging.info('Generate build files.')
            local_tarball_path = os.path.join(local_build_dir,
                                              'docker.tmp.tar.gz')
            self._wrap_dir_in_tarball(local_tarball_path, local_dir)
            # Upload to the context
            context = os.path.join(self._get_staging_location(),
                                   str(uuid.uuid4()) + '.tar.gz')
            GCSHelper.upload_gcs_file(local_tarball_path, context)

            # Run kaniko job
            kaniko_spec = self._generate_kaniko_spec(
                context=context,
                docker_filename=docker_filename,
                target_image=target_image)
            logging.info('Start a kaniko job for build.')
            from ._k8s_job_helper import K8sJobHelper
            k8s_helper = K8sJobHelper(self._k8s_client_configuration)
            result_pod_obj = k8s_helper.run_job(kaniko_spec, timeout)
            logging.info('Kaniko job complete.')

            # Clean up
            GCSHelper.remove_gcs_blob(context)

            # Returning image name with digest
            (image_repo, _, image_tag) = target_image.partition(':')
            # When Kaniko build completes successfully, the termination message is the hash digest of the newly built image. Otherwise it's empty. See https://github.com/GoogleContainerTools/kaniko#--digest-file https://kubernetes.io/docs/tasks/debug-application-cluster/determine-reason-pod-failure/#customizing-the-termination-message
            termination_message = [
                status.state.terminated.message
                for status in result_pod_obj.status.container_statuses
                if status.name == 'kaniko'
            ][0]  # Note: Using status.state instead of status.last_state since last_state entries can still be None
            image_digest = termination_message
            if not image_digest.startswith('sha256:'):
                raise RuntimeError(
                    "Kaniko returned invalid image digest: {}".format(
                        image_digest))
            strict_image_name = image_repo + '@' + image_digest
            logging.info(
                'Built and pushed image: {}.'.format(strict_image_name))
            return strict_image_name



# Copyright 2018-2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain
import os

import click
import kfp
from kfp import client
from kfp.cli import component
from kfp.cli import diagnose_me_cli
from kfp.cli import dsl
from kfp.cli import experiment
from kfp.cli import pipeline
from kfp.cli import recurring_run
from kfp.cli import run
from kfp.cli.output import OutputFormat
from kfp.cli.utils import aliased_plurals_group
from kfp.cli.utils import parsing

COMMANDS = {
    'client': {
        run.run, recurring_run.recurring_run, experiment.experiment,
        pipeline.pipeline
    },
    'no_client': {diagnose_me_cli.diagnose_me, component.component, dsl.dsl}
}

PROGRAM_NAME = 'kfp'

SHELL_FILES = {
    'bash': ['.bashrc'],
    'zsh': ['.zshrc'],
    'fish': ['.config', 'fish', 'completions', f'{PROGRAM_NAME}.fish']
}


def _create_completion(shell: str) -> str:
    return f'eval "$(_{PROGRAM_NAME.upper()}_COMPLETE={shell}_source {PROGRAM_NAME})"'


def _install_completion(shell: str) -> None:
    completion_statement = _create_completion(shell)
    source_file = os.path.join(os.path.expanduser('~'), *SHELL_FILES[shell])
    with open(source_file, 'a') as f:
        f.write('\n' + completion_statement + '\n')


@click.group(
    name=PROGRAM_NAME,
    cls=aliased_plurals_group.AliasedPluralsGroup,  # type: ignore
    commands=list(chain.from_iterable(COMMANDS.values())),  # type: ignore
    invoke_without_command=True)
@click.option(
    '--show-completion',
    type=click.Choice(list(SHELL_FILES.keys())),
    default=None)
@click.option(
    '--install-completion',
    type=click.Choice(list(SHELL_FILES.keys())),
    default=None)
@click.option('--endpoint', help=parsing.get_param_descr(client.Client, 'host'))
@click.option(
    '--iap-client-id', help=parsing.get_param_descr(client.Client, 'client_id'))
@click.option(
    '-n',
    '--namespace',
    default='kubeflow',
    show_default=True,
    help=parsing.get_param_descr(client.Client, 'namespace'))
@click.option(
    '--other-client-id',
    help=parsing.get_param_descr(client.Client, 'other_client_id'))
@click.option(
    '--other-client-secret',
    help=parsing.get_param_descr(client.Client, 'other_client_secret'))
@click.option(
    '--output',
    type=click.Choice(list(map(lambda x: x.name, OutputFormat))),
    default=OutputFormat.table.name,
    show_default=True,
    help='The formatting style for command output.')
@click.pass_context
@click.version_option(version=kfp.__version__, message='%(prog)s %(version)s')
def cli(ctx: click.Context, endpoint: str, iap_client_id: str, namespace: str,
        other_client_id: str, other_client_secret: str, output: OutputFormat,
        show_completion: str, install_completion: str):
    """Kubeflow Pipelines CLI."""
    if show_completion:
        click.echo(_create_completion(show_completion))
        return
    if install_completion:
        _install_completion(install_completion)
        return

    client_commands = set(
        chain.from_iterable([
            (command.name, f'{command.name}s')
            for command in COMMANDS['client']  # type: ignore
        ]))
    if ctx.invoked_subcommand not in client_commands:
        # Do not create a client for these subcommands
        return
    ctx.obj['client'] = client.Client(endpoint, iap_client_id, namespace,
                                      other_client_id, other_client_secret)
    ctx.obj['namespace'] = namespace
    ctx.obj['output'] = output



# Copyright 2019 The Kubeflow Authors. All Rights Reserved.
#
# Licensed under the Apache License,Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for collecting GCP related environment configurations."""

import enum
from typing import List, Optional, Text

from kfp.cli.diagnose_me import utility


class Commands(enum.Enum):
    """Enum for gcloud and gsutil commands."""
    GET_APIS = 1
    GET_CONTAINER_CLUSTERS = 2
    GET_CONTAINER_IMAGES = 3
    GET_DISKS = 4
    GET_GCLOUD_DEFAULT = 5
    GET_NETWORKS = 6
    GET_QUOTAS = 7
    GET_SCOPES = 8
    GET_SERVICE_ACCOUNTS = 9
    GET_STORAGE_BUCKETS = 10
    GET_GCLOUD_VERSION = 11
    GET_AUTH_LIST = 12


_command_string = {
    Commands.GET_APIS: 'services list',
    Commands.GET_CONTAINER_CLUSTERS: 'container clusters list',
    Commands.GET_CONTAINER_IMAGES: 'container images list',
    Commands.GET_DISKS: 'compute disks list',
    Commands.GET_GCLOUD_DEFAULT: 'config list --all',
    Commands.GET_NETWORKS: 'compute networks list',
    Commands.GET_QUOTAS: 'compute regions list',
    Commands.GET_SCOPES: 'compute instances list',
    Commands.GET_SERVICE_ACCOUNTS: 'iam service-accounts list',
    Commands.GET_STORAGE_BUCKETS: 'ls',
    Commands.GET_GCLOUD_VERSION: 'version',
    Commands.GET_AUTH_LIST: 'auth list',
}


def execute_gcloud_command(
        gcloud_command_list: List[Text],
        project_id: Optional[Text] = None,
        human_readable: Optional[bool] = False) -> utility.ExecutorResponse:
    """Function for invoking gcloud command.

    Args:
      gcloud_command_list: a command string list to be past to gcloud example
        format is ['config', 'list', '--all']
      project_id: specificies the project to run the commands against if not
        provided provided will use gcloud default project if one is configured
        otherwise will return an error message.
      human_readable: If false sets parameter --format json for all calls,
        otherwie output will be in human readable format.

    Returns:
      utility.ExecutorResponse with outputs from stdout,stderr and execution code.
    """
    command_list = ['gcloud']
    command_list.extend(gcloud_command_list)
    if not human_readable:
        command_list.extend(['--format', 'json'])

    if project_id is not None:
        command_list.extend(['--project', project_id])

    return utility.ExecutorResponse().execute_command(command_list)


def execute_gsutil_command(
        gsutil_command_list: List[Text],
        project_id: Optional[Text] = None) -> utility.ExecutorResponse:
    """Function for invoking gsutil command.

    This function takes in a gsutil parameter list and returns the results as a
    list of dictionaries.
    Args:
      gsutil_command_list: a command string list to be past to gsutil example
        format is ['config', 'list', '--all']
      project_id: specific project to check the QUOTASs for,if no project id is
        provided will use gcloud default project if one is configured otherwise
        will return an erro massage.

    Returns:
      utility.ExecutorResponse with outputs from stdout,stderr and execution code.
    """
    command_list = ['gsutil', *gsutil_command_list]
    if project_id is not None:
        command_list.extend(['-p', project_id])

    return utility.ExecutorResponse().execute_command(command_list)


def get_gcp_configuration(
        configuration: Commands,
        project_id: Optional[Text] = None,
        human_readable: Optional[bool] = False) -> utility.ExecutorResponse:
    """Captures the specified environment configuration.

    Captures the environment configuration for the specified setting such as
    NETWORKSing configuration, project QUOTASs, etc.

    Args:
      configuration: Commands for specific information to be retrieved
        - APIS:  Captures a complete list of enabled APISs and their configuration
          details under the specified project.
        - CONTAINER_CLUSTERS: List all visible k8 clusters under the project.
        - CONTAINER_IMAGES: List of all container images under the project
          container repo.
        - DISKS: List of storage allocated by the project including notebook
          instances as well as k8 pds with corresponding state.
        - GCLOUD_DEFAULT:  Environment default configuration for gcloud
        - NETWORKS: List all NETWORKSs and their configuration under the project.
        - QUOTAS:  Captures a complete list of  QUOTASs for project per
          region,returns the results as a list of dictionaries.
        - SCOPES: list of SCOPESs for each compute resources in the project.
        - SERVICE_ACCOUNTS: List of all service accounts that are enabled under
          this project.
        - STORAGE_BUCKETS: list of buckets and corresponding access information.
      project_id: specific project to check the QUOTASs for,if no project id is
        provided will use gcloud default project if one is configured otherwise
        will return an error message.
      human_readable: If true all output will be in human readable form insted of
        Json.

    Returns:
      A utility.ExecutorResponse with the output results for the specified
      command.
    """
    # storage bucket call requires execute_gsutil_command
    if configuration is Commands.GET_STORAGE_BUCKETS:
        return execute_gsutil_command(
            [_command_string[Commands.GET_STORAGE_BUCKETS]], project_id)

    # For all other cases can execute the command directly
    return execute_gcloud_command(_command_string[configuration].split(' '),
                                  project_id, human_readable)



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Optional

from kfp import client
from kfp.client import token_credentials_base
from kubernetes.client import configuration


class ServiceAccountTokenVolumeCredentials(
        token_credentials_base.TokenCredentialsBase):
    """Audience-bound ServiceAccountToken in the local filesystem.

    This is a credentials interface for audience-bound ServiceAccountTokens
    found in the local filesystem, that get refreshed by the kubelet.

    The constructor of the class expects a filesystem path.
    If not provided, it uses the path stored in the environment variable
    defined in ``client.KF_PIPELINES_SA_TOKEN_ENV``.
    If the environment variable is also empty, it falls back to the path
    specified in ``client.KF_PIPELINES_SA_TOKEN_PATH``.

    This method of authentication is meant for use inside a Kubernetes cluster.

    Relevant documentation:
    https://kubernetes.io/docs/tasks/configure-pod-container/configure-service-account/#service-account-token-volume-projection
    """

    def __init__(self, path: Optional[str] = None) -> None:
        """Constructs ServiceAccountTokenVolumeCredentials.

        Args:
            path (Optional[str], optional): Path of file containing auth token. Defaults to None.
        """
        self._token_path = (
            path or os.getenv(client.KF_PIPELINES_SA_TOKEN_ENV) or
            client.KF_PIPELINES_SA_TOKEN_PATH)

    def _get_token(self) -> str:
        """Reads the token found in file provided as `path` argument to
        ServiceAccountTokenVolumeCredentials constructor.

        Returns:
            str: The token.
        """
        token = None
        try:
            token = token_credentials_base.read_token_from_file(
                self._token_path)
        except OSError as e:
            logging.error("Failed to read a token from file '%s' (%s).",
                          self._token_path, str(e))
            raise
        return token

    def refresh_api_key_hook(self, config: configuration.Configuration) -> None:
        """Refreshes the api key using the token found in file provided as
        `path` argument to ServiceAccountTokenVolumeCredentials constructor.

        This is a helper function for registering token refresh with swagger
        generated clients.

        Args:
            config (kubernetes.client.configuration.Configuration):
                The configuration object that the client uses.

                The Configuration object of the kubernetes client's is the same
                with kfp_server_api.configuration.Configuration.
        """
        config.api_key['authorization'] = self._get_token()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfpmponents.for_loop."""
import unittest

from absl.testing import parameterized
from kfp.components import for_loop
from kfp.components import pipeline_channel


class ForLoopTest(parameterized.TestCase):

    @parameterized.parameters(
        {
            'collection_type': 'List[int]',
            'item_type': 'int',
        },
        {
            'collection_type': 'typing.List[str]',
            'item_type': 'str',
        },
        {
            'collection_type': 'typing.Tuple[  float   ]',
            'item_type': 'float',
        },
        {
            'collection_type': 'typing.Sequence[Dict[str, str]]',
            'item_type': 'Dict[str, str]',
        },
        {
            'collection_type': 'List',
            'item_type': None,
        },
    )
    def test_get_loop_item_type(self, collection_type, item_type):
        self.assertEqual(
            for_loop._get_loop_item_type(collection_type), item_type)

    @parameterized.parameters(
        {
            'dict_type': 'Dict[str, int]',
            'value_type': 'int',
        },
        {
            'dict_type': 'typing.Mapping[str,float]',
            'value_type': 'float',
        },
        {
            'dict_type': 'typing.Mapping[str, Dict[str, str] ]',
            'value_type': 'Dict[str, str]',
        },
        {
            'dict_type': 'dict',
            'value_type': None,
        },
    )
    def test_get_subvar_type(self, dict_type, value_type):
        self.assertEqual(for_loop._get_subvar_type(dict_type), value_type)

    @parameterized.parameters(
        {
            'channel':
                pipeline_channel.PipelineParameterChannel(
                    name='param1',
                    channel_type='List[str]',
                ),
            'expected_serialization_value':
                '{{channel:task=;name=param1-loop-item;type=str;}}',
        },
        {
            'channel':
                pipeline_channel.PipelineParameterChannel(
                    name='output1',
                    channel_type='List[Dict[str, str]]',
                    task_name='task1',
                ),
            'expected_serialization_value':
                '{{channel:task=task1;name=output1-loop-item;type=Dict[str, str];}}',
        },
    )
    def test_loop_argument_from_pipeline_channel(self, channel,
                                                 expected_serialization_value):
        loop_argument = for_loop.LoopArgument.from_pipeline_channel(channel)
        self.assertEqual(loop_argument.items_or_pipeline_channel, channel)
        self.assertEqual(str(loop_argument), expected_serialization_value)

    @parameterized.parameters(
        {
            'raw_items': ['a', 'b', 'c'],
            'name_code':
                '1',
            'expected_serialization_value':
                '{{channel:task=;name=loop-item-param-1;type=str;}}',
        },
        {
            'raw_items': [
                {
                    'A_a': 1
                },
                {
                    'A_a': 2
                },
            ],
            'name_code':
                '2',
            'expected_serialization_value':
                '{{channel:task=;name=loop-item-param-2;type=dict;}}',
        },
    )
    def test_loop_argument_from_raw_items(self, raw_items, name_code,
                                          expected_serialization_value):
        loop_argument = for_loop.LoopArgument.from_raw_items(
            raw_items, name_code)
        self.assertEqual(loop_argument.items_or_pipeline_channel, raw_items)
        self.assertEqual(str(loop_argument), expected_serialization_value)

    @parameterized.parameters(
        {
            'name': 'abc-loop-item',
            'expected_result': True
        },
        {
            'name': 'abc-loop-item-subvar-a',
            'expected_result': True
        },
        {
            'name': 'loop-item-param-1',
            'expected_result': True
        },
        {
            'name': 'loop-item-param-1-subvar-a',
            'expected_result': True
        },
        {
            'name': 'param1',
            'expected_result': False
        },
    )
    def test_name_is_loop_argument(self, name, expected_result):
        self.assertEqual(
            for_loop.LoopArgument.name_is_loop_argument(name), expected_result)

    @parameterized.parameters(
        {
            'subvar_name': 'a',
            'valid': True
        },
        {
            'subvar_name': 'A_a',
            'valid': True
        },
        {
            'subvar_name': 'a0',
            'valid': True
        },
        {
            'subvar_name': 'a-b',
            'valid': False
        },
        {
            'subvar_name': '0',
            'valid': False
        },
        {
            'subvar_name': 'a#',
            'valid': False
        },
    )
    def test_create_loop_argument_varaible(self, subvar_name, valid):
        loop_argument = for_loop.LoopArgument.from_pipeline_channel(
            pipeline_channel.PipelineParameterChannel(
                name='param1',
                channel_type='List[Dict[str, str]]',
            ))
        if valid:
            loop_arg_var = for_loop.LoopArgumentVariable(
                loop_argument=loop_argument,
                subvar_name=subvar_name,
            )
            self.assertEqual(loop_arg_var.loop_argument, loop_argument)
            self.assertEqual(loop_arg_var.subvar_name, subvar_name)
        else:
            with self.assertRaisesRegex(ValueError,
                                        'Tried to create subvariable'):
                for_loop.LoopArgumentVariable(
                    loop_argument=loop_argument,
                    subvar_name=subvar_name,
                )


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains tests for kfp.components.placeholders."""
import os
import tempfile
from typing import Any, List

from absl.testing import parameterized
from kfp import compiler
from kfp import dsl
from kfp.components import placeholders
from kfp.dsl import Artifact
from kfp.dsl import Dataset
from kfp.dsl import Input
from kfp.dsl import Output


class TestExecutorInputPlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(placeholders.ExecutorInputPlaceholder()._to_string(),
                         '{{$}}')


class TestInputValuePlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(
            placeholders.InputValuePlaceholder('input1')._to_string(),
            "{{$.inputs.parameters['input1']}}")


class TestInputPathPlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(
            placeholders.InputPathPlaceholder('input1')._to_string(),
            "{{$.inputs.artifacts['input1'].path}}")


class TestInputUriPlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(
            placeholders.InputUriPlaceholder('input1')._to_string(),
            "{{$.inputs.artifacts['input1'].uri}}")


class TestInputMetadataPlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(
            placeholders.InputMetadataPlaceholder('input1')._to_string(),
            "{{$.inputs.artifacts['input1'].metadata}}")

    def test_access_top_level_metadata_key(self):

        @dsl.container_component
        def echo(d: Input[Dataset]):
            return dsl.ContainerSpec(
                image='alpine', command=['echo', d.metadata['key']])

        self.assertEqual(
            echo.pipeline_spec.deployment_spec['executors']['exec-echo']
            ['container']['command'][1],
            "{{$.inputs.artifacts['d'].metadata['key']}}")


class TestOutputPathPlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(
            placeholders.OutputPathPlaceholder('output1')._to_string(),
            "{{$.outputs.artifacts['output1'].path}}")


class TestOutputParameterPlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(
            placeholders.OutputParameterPlaceholder('output1')._to_string(),
            "{{$.outputs.parameters['output1'].output_file}}")


class TestOutputUriPlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(
            placeholders.OutputUriPlaceholder('output1')._to_string(),
            "{{$.outputs.artifacts['output1'].uri}}")


class TestOutputMetadataPlaceholder(parameterized.TestCase):

    def test(self):
        self.assertEqual(
            placeholders.OutputMetadataPlaceholder('output1')._to_string(),
            "{{$.outputs.artifacts['output1'].metadata}}")

    def test_access_top_level_metadata_key(self):

        @dsl.container_component
        def echo(d: Output[Dataset]):
            return dsl.ContainerSpec(
                image='alpine', command=['echo', d.metadata['key']])

        self.assertEqual(
            echo.pipeline_spec.deployment_spec['executors']['exec-echo']
            ['container']['command'][1],
            "{{$.outputs.artifacts['d'].metadata['key']}}")


class TestIfPresentPlaceholder(parameterized.TestCase):

    @parameterized.parameters([
        (placeholders.IfPresentPlaceholder(
            input_name='input1', then=['then'], else_=['something']),
         '{"IfPresent": {"InputName": "input1", "Then": ["then"], "Else": ["something"]}}'
        ),
        (placeholders.IfPresentPlaceholder(
            input_name='input1', then='then', else_='something'),
         '{"IfPresent": {"InputName": "input1", "Then": "then", "Else": "something"}}'
        ),
        (placeholders.IfPresentPlaceholder(
            input_name='input1', then='then', else_=['something']),
         '{"IfPresent": {"InputName": "input1", "Then": "then", "Else": ["something"]}}'
        ),
        (placeholders.IfPresentPlaceholder(
            input_name='input1', then=['then'], else_='something'),
         '{"IfPresent": {"InputName": "input1", "Then": ["then"], "Else": "something"}}'
        ),
    ])
    def test_strings_and_lists(
            self, placeholder_obj: placeholders.IfPresentPlaceholder,
            placeholder: str):
        self.assertEqual(placeholder_obj._to_string(), placeholder)

    @parameterized.parameters([
        (placeholders.IfPresentPlaceholder(
            input_name='input1',
            then=[
                '--flag',
                placeholders.OutputUriPlaceholder(output_name='output1')
            ],
            else_=[
                '--flag',
                placeholders.OutputMetadataPlaceholder(output_name='output1')
            ]),
         """{"IfPresent": {"InputName": "input1", "Then": ["--flag", "{{$.outputs.artifacts['output1'].uri}}"], "Else": ["--flag", "{{$.outputs.artifacts['output1'].metadata}}"]}}"""
        ),
        (placeholders.IfPresentPlaceholder(
            input_name='input1',
            then=placeholders.InputPathPlaceholder(input_name='input2'),
            else_=placeholders.InputValuePlaceholder(input_name='input2')),
         """{"IfPresent": {"InputName": "input1", "Then": "{{$.inputs.artifacts['input2'].path}}", "Else": "{{$.inputs.parameters['input2']}}"}}"""
        ),
    ])
    def test_with_primitive_placeholders(
            self, placeholder_obj: placeholders.IfPresentPlaceholder,
            placeholder: str):
        self.assertEqual(placeholder_obj._to_string(), placeholder)

    def test_if_present_with_single_element_simple_can_be_compiled(self):

        @dsl.container_component
        def container_component(a: str):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    placeholders.IfPresentPlaceholder(
                        input_name='a', then='b', else_='c')
                ])

        with tempfile.TemporaryDirectory() as tempdir:
            output_yaml = os.path.join(tempdir, 'component.yaml')
            compiler.Compiler().compile(
                pipeline_func=container_component, package_path=output_yaml)

    def test_if_present_with_single_element_parameter_reference_can_be_compiled(
            self):

        @dsl.container_component
        def container_component(a: str):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    placeholders.IfPresentPlaceholder(
                        input_name='a', then=a, else_='c')
                ])

        with tempfile.TemporaryDirectory() as tempdir:
            output_yaml = os.path.join(tempdir, 'component.yaml')
            compiler.Compiler().compile(
                pipeline_func=container_component, package_path=output_yaml)

    def test_if_present_with_single_element_artifact_reference_can_be_compiled(
            self):

        @dsl.container_component
        def container_component(a: dsl.Input[dsl.Artifact]):
            return dsl.ContainerSpec(
                image='alpine',
                command=[
                    placeholders.IfPresentPlaceholder(
                        input_name='a', then=a.path, else_='c')
                ])

        with tempfile.TemporaryDirectory() as tempdir:
            output_yaml = os.path.join(tempdir, 'component.yaml')
            compiler.Compiler().compile(
                pipeline_func=container_component, package_path=output_yaml)


class TestConcatPlaceholder(parameterized.TestCase):

    @parameterized.parameters([
        (placeholders.ConcatPlaceholder(['a']), '{"Concat": ["a"]}'),
        (placeholders.ConcatPlaceholder(['a', 'b']), '{"Concat": ["a", "b"]}'),
    ])
    def test_strings(self, placeholder_obj: placeholders.ConcatPlaceholder,
                     placeholder_string: str):
        self.assertEqual(placeholder_obj._to_string(), placeholder_string)

    @parameterized.parameters([
        (placeholders.ConcatPlaceholder([
            'a', placeholders.InputPathPlaceholder(input_name='input2')
        ]), """{"Concat": ["a", "{{$.inputs.artifacts['input2'].path}}"]}"""),
        (placeholders.ConcatPlaceholder([
            placeholders.InputValuePlaceholder(input_name='input2'), 'b'
        ]), """{"Concat": ["{{$.inputs.parameters['input2']}}", "b"]}"""),
    ])
    def test_primitive_placeholders(
            self, placeholder_obj: placeholders.ConcatPlaceholder,
            placeholder_string: str):
        self.assertEqual(placeholder_obj._to_string(), placeholder_string)


class TestContainerPlaceholdersTogether(parameterized.TestCase):

    @parameterized.parameters([
        (placeholders.ConcatPlaceholder([
            'a', placeholders.ConcatPlaceholder(['b', 'c'])
        ]), '{"Concat": ["a", {"Concat": ["b", "c"]}]}'),
        (placeholders.ConcatPlaceholder([
            'a',
            placeholders.ConcatPlaceholder([
                'b',
                placeholders.ConcatPlaceholder([
                    'c',
                    placeholders.InputValuePlaceholder(input_name='input2')
                ])
            ])
        ]),
         """{"Concat": ["a", {"Concat": ["b", {"Concat": ["c", "{{$.inputs.parameters['input2']}}"]}]}]}"""
        ),
        (placeholders.ConcatPlaceholder([
            'a',
            placeholders.ConcatPlaceholder([
                'b',
                placeholders.IfPresentPlaceholder(
                    input_name='output1', then='then', else_='something')
            ])
        ]),
         '{"Concat": ["a", {"Concat": ["b", {"IfPresent": {"InputName": "output1", "Then": "then", "Else": "something"}}]}]}'
        ),
        (placeholders.ConcatPlaceholder([
            'a',
            placeholders.IfPresentPlaceholder(
                input_name='output1',
                then=placeholders.ConcatPlaceholder([
                    '--',
                    'flag',
                    placeholders.InputPathPlaceholder(input_name='input2'),
                ]),
                else_='b'),
            'c',
        ]),
         """{"Concat": ["a", {"IfPresent": {"InputName": "output1", "Then": {"Concat": ["--", "flag", "{{$.inputs.artifacts['input2'].path}}"]}, "Else": "b"}}, "c"]}"""
        ),
        (placeholders.ConcatPlaceholder([
            'a',
            placeholders.IfPresentPlaceholder(
                input_name='output1',
                then=placeholders.ConcatPlaceholder(['--', 'flag']),
                else_=placeholders.InputPathPlaceholder(input_name='input2')),
            'c',
        ]),
         """{"Concat": ["a", {"IfPresent": {"InputName": "output1", "Then": {"Concat": ["--", "flag"]}, "Else": "{{$.inputs.artifacts['input2'].path}}"}}, "c"]}"""
        ),
    ])
    def test_valid(self, placeholder_obj: placeholders.IfPresentPlaceholder,
                   placeholder: str):
        self.assertEqual(placeholder_obj._to_string(), placeholder)

    def test_only_single_element_ifpresent_inside_concat_outer(self):
        with self.assertRaisesRegex(
                ValueError,
                f'Please use a single element for `then` and `else_` only\.'):
            placeholders.ConcatPlaceholder([
                'b',
                placeholders.IfPresentPlaceholder(
                    input_name='output1', then=['then'], else_=['something'])
            ])

    def test_only_single_element_ifpresent_inside_concat_recursive(self):
        with self.assertRaisesRegex(
                ValueError,
                f'Please use a single element for `then` and `else_` only\.'):
            placeholders.ConcatPlaceholder([
                'a',
                placeholders.ConcatPlaceholder([
                    'b',
                    placeholders.IfPresentPlaceholder(
                        input_name='output1',
                        then=['then'],
                        else_=['something'])
                ])
            ])

        with self.assertRaisesRegex(
                ValueError,
                f'Please use a single element for `then` and `else_` only\.'):
            placeholders.ConcatPlaceholder([
                'a',
                placeholders.ConcatPlaceholder([
                    'b',
                    placeholders.IfPresentPlaceholder(
                        input_name='output1',
                        then=placeholders.ConcatPlaceholder([
                            placeholders.IfPresentPlaceholder(
                                input_name='a', then=['b'])
                        ]),
                        else_='something')
                ])
            ])

    def test_only_single_element_in_nested_ifpresent_inside_concat(self):
        with self.assertRaisesRegex(
                ValueError,
                f'Please use a single element for `then` and `else_` only\.'):
            dsl.ConcatPlaceholder([
                'my-prefix-',
                dsl.IfPresentPlaceholder(
                    input_name='input1',
                    then=[
                        dsl.IfPresentPlaceholder(
                            input_name='input1',
                            then=dsl.ConcatPlaceholder(['infix-', 'value']))
                    ])
            ])

    def test_recursive_nested_placeholder_validation_does_not_exit_when_first_valid_then_is_found(
            self):
        with self.assertRaisesRegex(
                ValueError,
                f'Please use a single element for `then` and `else_` only\.'):
            dsl.ConcatPlaceholder([
                'my-prefix-',
                dsl.IfPresentPlaceholder(
                    input_name='input1',
                    then=dsl.IfPresentPlaceholder(
                        input_name='input1',
                        then=[dsl.ConcatPlaceholder(['infix-', 'value'])]))
            ])

    def test_only_single_element_in_nested_ifpresent_inside_concat_with_outer_ifpresent(
            self):
        with self.assertRaisesRegex(
                ValueError,
                f'Please use a single element for `then` and `else_` only\.'):
            dsl.IfPresentPlaceholder(
                input_name='input_1',
                then=dsl.ConcatPlaceholder([
                    'my-prefix-',
                    dsl.IfPresentPlaceholder(
                        input_name='input1',
                        then=dsl.IfPresentPlaceholder(
                            input_name='input1',
                            then=[dsl.ConcatPlaceholder(['infix-', 'value'])]))
                ]))

    def test_valid_then_but_invalid_else(self):
        with self.assertRaisesRegex(
                ValueError,
                f'Please use a single element for `then` and `else_` only\.'):
            dsl.ConcatPlaceholder([
                'my-prefix-',
                dsl.IfPresentPlaceholder(
                    input_name='input1',
                    then=dsl.IfPresentPlaceholder(
                        input_name='input1',
                        then='single-element',
                        else_=['one', 'two']))
            ])


class TestListOfArtifactsInContainerComponentPlaceholders(
        parameterized.TestCase):

    def test_compile_component1(self):

        @dsl.container_component
        def comp(input_list: Input[List[Artifact]]):
            return dsl.ContainerSpec(
                image='alpine', command=[input_list], args=[input_list])

        self.assertEqual(
            comp.pipeline_spec.deployment_spec['executors']['exec-comp']
            ['container']['command'][0], "{{$.inputs.artifacts['input_list']}}")
        self.assertEqual(
            comp.pipeline_spec.deployment_spec['executors']['exec-comp']
            ['container']['args'][0], "{{$.inputs.artifacts['input_list']}}")

    def test_compile_component2(self):

        @dsl.container_component
        def comp(new_name: Input[List[Dataset]]):
            return dsl.ContainerSpec(
                image='alpine', command=[new_name], args=[new_name])

        self.assertEqual(
            comp.pipeline_spec.deployment_spec['executors']['exec-comp']
            ['container']['command'][0], "{{$.inputs.artifacts['new_name']}}")
        self.assertEqual(
            comp.pipeline_spec.deployment_spec['executors']['exec-comp']
            ['container']['args'][0], "{{$.inputs.artifacts['new_name']}}")

    def test_cannot_access_name(self):
        with self.assertRaisesRegex(AttributeError,
                                    'Cannot access an attribute'):

            @dsl.container_component
            def comp(new_name: Input[List[Dataset]]):
                return dsl.ContainerSpec(
                    image='alpine', command=[new_name.name])

    def test_cannot_access_uri(self):
        with self.assertRaisesRegex(AttributeError,
                                    'Cannot access an attribute'):

            @dsl.container_component
            def comp(new_name: Input[List[Dataset]]):
                return dsl.ContainerSpec(image='alpine', command=[new_name.uri])

    def test_cannot_access_metadata(self):
        with self.assertRaisesRegex(AttributeError,
                                    'Cannot access an attribute'):

            @dsl.container_component
            def comp(new_name: Input[List[Dataset]]):
                return dsl.ContainerSpec(
                    image='alpine', command=[new_name.metadata])

    def test_cannot_access_path(self):
        with self.assertRaisesRegex(AttributeError,
                                    'Cannot access an attribute'):

            @dsl.container_component
            def comp(new_name: Input[List[Dataset]]):
                return dsl.ContainerSpec(
                    image='alpine', command=[new_name.path])

    def test_cannot_access_individual_artifact(self):
        with self.assertRaisesRegex(KeyError,
                                    'Cannot access individual artifacts'):

            @dsl.container_component
            def comp(new_name: Input[List[Dataset]]):
                return dsl.ContainerSpec(image='alpine', command=[new_name[0]])


class TestConvertCommandLineElementToStringOrStruct(parameterized.TestCase):

    @parameterized.parameters(['a', 'word', 1])
    def test_pass_through(self, val: Any):
        self.assertEqual(
            placeholders.convert_command_line_element_to_string_or_struct(val),
            val)

    @parameterized.parameters([
        (placeholders.ExecutorInputPlaceholder(), '{{$}}'),
        (placeholders.InputValuePlaceholder('input1'),
         """{{$.inputs.parameters['input1']}}"""),
        (placeholders.OutputPathPlaceholder('output1'),
         """{{$.outputs.artifacts['output1'].path}}"""),
    ])
    def test_primitive_placeholder(self, placeholder: Any, expected: str):
        self.assertEqual(
            placeholders.convert_command_line_element_to_string_or_struct(
                placeholder), expected)


class OtherPlaceholderTests(parameterized.TestCase):

    def test_primitive_placeholder_can_be_used_in_fstring1(self):

        @dsl.container_component
        def echo_bool(boolean: bool = True):
            return dsl.ContainerSpec(
                image='alpine', command=['sh', '-c', f'echo {boolean}'])

        self.assertEqual(
            echo_bool.component_spec.implementation.container.command,
            ['sh', '-c', "echo {{$.inputs.parameters['boolean']}}"])

    def test_primitive_placeholder_can_be_used_in_fstring2(self):

        @dsl.container_component
        def container_with_placeholder_in_fstring(
            output_artifact: Output[Artifact],
            text1: str,
        ):
            return dsl.ContainerSpec(
                image='python:3.7',
                command=[
                    'my_program',
                    f'prefix-{text1}',
                    f'{output_artifact.uri}/0',
                ])

        self.assertEqual(
            container_with_placeholder_in_fstring.component_spec.implementation
            .container.command, [
                'my_program',
                "prefix-{{$.inputs.parameters['text1']}}",
                "{{$.outputs.artifacts['output_artifact'].uri}}/0",
            ])

    def test_cannot_use_concat_placeholder_in_f_string(self):

        with self.assertRaisesRegex(
                ValueError, 'Cannot use ConcatPlaceholder in an f-string.'):

            @dsl.container_component
            def container_with_placeholder_in_fstring(
                text1: str,
                text2: str,
            ):
                return dsl.ContainerSpec(
                    image='python:3.7',
                    command=[
                        'my_program',
                        f'another-prefix-{dsl.ConcatPlaceholder([text1, text2])}',
                    ])

    def test_cannot_use_ifpresent_placeholder_in_f_string(self):

        with self.assertRaisesRegex(
                ValueError, 'Cannot use IfPresentPlaceholder in an f-string.'):

            @dsl.container_component
            def container_with_placeholder_in_fstring(
                text1: str,
                text2: str,
            ):
                return dsl.ContainerSpec(
                    image='python:3.7',
                    command=[
                        'echo',
                        f"another-prefix-{dsl.IfPresentPlaceholder(input_name='text1', then=['val'])}",
                    ])



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for kfp.compiler.pipeline_spec_builder."""

import unittest

from absl.testing import parameterized
from google.protobuf import json_format
from google.protobuf import struct_pb2
from kfp.compiler import pipeline_spec_builder
from kfp.pipeline_spec import pipeline_spec_pb2
import yaml


class PipelineSpecBuilderTest(parameterized.TestCase):

    def setUp(self):
        self.maxDiff = None

    @parameterized.parameters(
        {
            'parameter_type': pipeline_spec_pb2.ParameterType.NUMBER_INTEGER,
            'default_value': None,
            'expected': struct_pb2.Value(),
        },
        {
            'parameter_type': pipeline_spec_pb2.ParameterType.NUMBER_INTEGER,
            'default_value': 1,
            'expected': struct_pb2.Value(number_value=1),
        },
        {
            'parameter_type': pipeline_spec_pb2.ParameterType.NUMBER_DOUBLE,
            'default_value': 1.2,
            'expected': struct_pb2.Value(number_value=1.2),
        },
        {
            'parameter_type': pipeline_spec_pb2.ParameterType.STRING,
            'default_value': 'text',
            'expected': struct_pb2.Value(string_value='text'),
        },
        {
            'parameter_type': pipeline_spec_pb2.ParameterType.BOOLEAN,
            'default_value': True,
            'expected': struct_pb2.Value(bool_value=True),
        },
        {
            'parameter_type': pipeline_spec_pb2.ParameterType.BOOLEAN,
            'default_value': False,
            'expected': struct_pb2.Value(bool_value=False),
        },
        {
            'parameter_type':
                pipeline_spec_pb2.ParameterType.STRUCT,
            'default_value': {
                'a': 1,
                'b': 2,
            },
            'expected':
                struct_pb2.Value(
                    struct_value=struct_pb2.Struct(
                        fields={
                            'a': struct_pb2.Value(number_value=1),
                            'b': struct_pb2.Value(number_value=2),
                        })),
        },
        {
            'parameter_type':
                pipeline_spec_pb2.ParameterType.LIST,
            'default_value': ['a', 'b'],
            'expected':
                struct_pb2.Value(
                    list_value=struct_pb2.ListValue(values=[
                        struct_pb2.Value(string_value='a'),
                        struct_pb2.Value(string_value='b'),
                    ])),
        },
        {
            'parameter_type':
                pipeline_spec_pb2.ParameterType.LIST,
            'default_value': [{
                'a': 1,
                'b': 2
            }, {
                'a': 10,
                'b': 20
            }],
            'expected':
                struct_pb2.Value(
                    list_value=struct_pb2.ListValue(values=[
                        struct_pb2.Value(
                            struct_value=struct_pb2.Struct(
                                fields={
                                    'a': struct_pb2.Value(number_value=1),
                                    'b': struct_pb2.Value(number_value=2),
                                })),
                        struct_pb2.Value(
                            struct_value=struct_pb2.Struct(
                                fields={
                                    'a': struct_pb2.Value(number_value=10),
                                    'b': struct_pb2.Value(number_value=20),
                                })),
                    ])),
        },
    )
    def test_fill_in_component_input_default_value(self, parameter_type,
                                                   default_value, expected):
        component_spec = pipeline_spec_pb2.ComponentSpec(
            input_definitions=pipeline_spec_pb2.ComponentInputsSpec(
                parameters={
                    'input1':
                        pipeline_spec_pb2.ComponentInputsSpec.ParameterSpec(
                            parameter_type=parameter_type)
                }))
        pipeline_spec_builder._fill_in_component_input_default_value(
            component_spec=component_spec,
            input_name='input1',
            default_value=default_value)

        self.assertEqual(
            expected,
            component_spec.input_definitions.parameters['input1'].default_value,
        )

    def test_merge_deployment_spec_and_component_spec(self):
        main_deployment_config = pipeline_spec_pb2.PipelineDeploymentConfig()
        main_deployment_config.executors['exec-1'].CopyFrom(
            pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
                container=pipeline_spec_pb2.PipelineDeploymentConfig
                .PipelineContainerSpec(image='img-1', command=['cmd'])))
        main_pipeline_spec = pipeline_spec_pb2.PipelineSpec()
        main_pipeline_spec.components['comp-1'].CopyFrom(
            pipeline_spec_pb2.ComponentSpec(executor_label='exec-1'))

        sub_deployment_config = pipeline_spec_pb2.PipelineDeploymentConfig()
        sub_deployment_config.executors['exec-1'].CopyFrom(
            pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
                container=pipeline_spec_pb2.PipelineDeploymentConfig
                .PipelineContainerSpec(image='img-a', command=['cmd'])))
        sub_deployment_config.executors['exec-2'].CopyFrom(
            pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
                container=pipeline_spec_pb2.PipelineDeploymentConfig
                .PipelineContainerSpec(image='img-b', command=['cmd'])))

        sub_pipeline_spec = pipeline_spec_pb2.PipelineSpec()
        sub_pipeline_spec.deployment_spec.update(
            json_format.MessageToDict(sub_deployment_config))
        sub_pipeline_spec.components['comp-1'].CopyFrom(
            pipeline_spec_pb2.ComponentSpec(executor_label='exec-1'))
        sub_pipeline_spec.components['comp-2'].CopyFrom(
            pipeline_spec_pb2.ComponentSpec(executor_label='exec-2'))
        sub_pipeline_spec.root.CopyFrom(
            pipeline_spec_pb2.ComponentSpec(dag=pipeline_spec_pb2.DagSpec()))
        sub_pipeline_spec.root.dag.tasks['task-1'].CopyFrom(
            pipeline_spec_pb2.PipelineTaskSpec(
                component_ref=pipeline_spec_pb2.ComponentRef(name='comp-1')))
        sub_pipeline_spec.root.dag.tasks['task-2'].CopyFrom(
            pipeline_spec_pb2.PipelineTaskSpec(
                component_ref=pipeline_spec_pb2.ComponentRef(name='comp-2')))

        expected_sub_pipeline_spec = pipeline_spec_pb2.PipelineSpec.FromString(
            sub_pipeline_spec.SerializeToString())

        expected_main_deployment_config = pipeline_spec_pb2.PipelineDeploymentConfig(
        )
        expected_main_deployment_config.executors['exec-1'].CopyFrom(
            pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
                container=pipeline_spec_pb2.PipelineDeploymentConfig
                .PipelineContainerSpec(image='img-1', command=['cmd'])))
        expected_main_deployment_config.executors['exec-1-2'].CopyFrom(
            pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
                container=pipeline_spec_pb2.PipelineDeploymentConfig
                .PipelineContainerSpec(image='img-a', command=['cmd'])))
        expected_main_deployment_config.executors['exec-2'].CopyFrom(
            pipeline_spec_pb2.PipelineDeploymentConfig.ExecutorSpec(
                container=pipeline_spec_pb2.PipelineDeploymentConfig
                .PipelineContainerSpec(image='img-b', command=['cmd'])))

        expected_main_pipeline_spec = pipeline_spec_pb2.PipelineSpec()
        expected_main_pipeline_spec.components['comp-1'].CopyFrom(
            pipeline_spec_pb2.ComponentSpec(executor_label='exec-1'))
        expected_main_pipeline_spec.components['comp-1-2'].CopyFrom(
            pipeline_spec_pb2.ComponentSpec(executor_label='exec-1-2'))
        expected_main_pipeline_spec.components['comp-2'].CopyFrom(
            pipeline_spec_pb2.ComponentSpec(executor_label='exec-2'))

        expected_sub_pipeline_spec_root = pipeline_spec_pb2.ComponentSpec(
            dag=pipeline_spec_pb2.DagSpec())
        expected_sub_pipeline_spec_root.dag.tasks['task-1'].CopyFrom(
            pipeline_spec_pb2.PipelineTaskSpec(
                component_ref=pipeline_spec_pb2.ComponentRef(name='comp-1-2')))
        expected_sub_pipeline_spec_root.dag.tasks['task-2'].CopyFrom(
            pipeline_spec_pb2.PipelineTaskSpec(
                component_ref=pipeline_spec_pb2.ComponentRef(name='comp-2')))

        sub_pipeline_spec_copy, sub_platform_spec = pipeline_spec_builder.merge_deployment_spec_and_component_spec(
            main_pipeline_spec=main_pipeline_spec,
            main_deployment_config=main_deployment_config,
            sub_pipeline_spec=sub_pipeline_spec,
            main_platform_spec=pipeline_spec_pb2.PlatformSpec(),
            sub_platform_spec=pipeline_spec_pb2.PlatformSpec(),
        )

        self.assertEqual(sub_pipeline_spec, expected_sub_pipeline_spec)
        self.assertEqual(main_pipeline_spec, expected_main_pipeline_spec)
        self.assertEqual(main_deployment_config,
                         expected_main_deployment_config)
        self.assertEqual(sub_pipeline_spec_copy.root,
                         expected_sub_pipeline_spec_root)


class TestPlatformConfigToPlatformSpec(unittest.TestCase):

    def test(self):
        platform_config = {
            'platform_key': {
                'feat1': [1, 2, 3],
                'feat2': 'hello',
                'feat3': {
                    'k': 'v'
                }
            }
        }
        actual = pipeline_spec_builder.platform_config_to_platform_spec(
            platform_config=platform_config, executor_label='exec-comp')
        expected = pipeline_spec_pb2.PlatformSpec()
        json_format.ParseDict(
            {
                'platforms': {
                    'platform_key': {
                        'deployment_spec': {
                            'executors': {
                                'exec-comp': {
                                    'feat1': [1, 2, 3],
                                    'feat2': 'hello',
                                    'feat3': {
                                        'k': 'v'
                                    }
                                }
                            }
                        }
                    }
                }
            }, expected)
        self.assertEqual(actual, expected)


class TestMergePlatformSpecs(unittest.TestCase):

    def test_merge_different_executors(self):
        main_spec = pipeline_spec_pb2.PlatformSpec()
        json_format.ParseDict(
            {
                'platforms': {
                    'platform_key': {
                        'deployment_spec': {
                            'executors': {
                                'exec-comp': {
                                    'feat1': [1, 2, 3],
                                    'feat2': 'hello',
                                    'feat3': {
                                        'k': 'v'
                                    }
                                }
                            }
                        }
                    }
                }
            }, main_spec)

        sub_spec = pipeline_spec_pb2.PlatformSpec()
        json_format.ParseDict(
            {
                'platforms': {
                    'platform_key': {
                        'deployment_spec': {
                            'executors': {
                                'exec-other': {
                                    'feat1': [4, 5, 6],
                                    'feat2': 'goodbye',
                                    'feat3': {
                                        'k2': 'v2'
                                    }
                                }
                            }
                        }
                    }
                }
            }, sub_spec)

        expected = pipeline_spec_pb2.PlatformSpec()
        json_format.ParseDict(
            {
                'platforms': {
                    'platform_key': {
                        'deployment_spec': {
                            'executors': {
                                'exec-other': {
                                    'feat1': [4, 5, 6],
                                    'feat2': 'goodbye',
                                    'feat3': {
                                        'k2': 'v2'
                                    }
                                },
                                'exec-comp': {
                                    'feat1': [1, 2, 3],
                                    'feat2': 'hello',
                                    'feat3': {
                                        'k': 'v'
                                    }
                                }
                            }
                        }
                    }
                }
            }, expected)

        pipeline_spec_builder.merge_platform_specs(main_spec, sub_spec)
        self.assertEqual(main_spec, expected)

    def test_merge_different_platforms(self):
        base_spec = pipeline_spec_pb2.PlatformSpec()
        json_format.ParseDict(
            {
                'platforms': {
                    'platform1': {
                        'deployment_spec': {
                            'executors': {
                                'exec-comp': {
                                    'feat1': [1, 2, 3],
                                    'feat2': 'hello',
                                    'feat3': {
                                        'k': 'v'
                                    }
                                }
                            }
                        }
                    }
                }
            }, base_spec)

        sub_spec = pipeline_spec_pb2.PlatformSpec()
        json_format.ParseDict(
            {
                'platforms': {
                    'platform2': {
                        'deployment_spec': {
                            'executors': {
                                'exec-comp': {
                                    'feat1': [4, 5, 6],
                                    'feat2': 'goodbye',
                                    'feat3': {
                                        'k2': 'v2'
                                    }
                                }
                            }
                        }
                    }
                }
            }, sub_spec)

        expected = pipeline_spec_pb2.PlatformSpec()
        json_format.ParseDict(
            {
                'platforms': {
                    'platform1': {
                        'deployment_spec': {
                            'executors': {
                                'exec-comp': {
                                    'feat1': [1, 2, 3],
                                    'feat2': 'hello',
                                    'feat3': {
                                        'k': 'v'
                                    }
                                }
                            }
                        }
                    },
                    'platform2': {
                        'deployment_spec': {
                            'executors': {
                                'exec-comp': {
                                    'feat1': [4, 5, 6],
                                    'feat2': 'goodbye',
                                    'feat3': {
                                        'k2': 'v2'
                                    }
                                }
                            }
                        }
                    }
                }
            }, expected)

        pipeline_spec_builder.merge_platform_specs(base_spec, sub_spec)
        self.assertEqual(base_spec, expected)


def pipeline_spec_from_file(filepath: str) -> str:
    with open(filepath, 'r') as f:
        dictionary = yaml.safe_load(f)
    return json_format.ParseDict(dictionary, pipeline_spec_pb2.PipelineSpec())


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
import subprocess
import time
import json
import click
import shutil
import datetime
from typing import List

import kfp_server_api
from kfp.deprecated._client import Client
from kfp.deprecated.cli.output import print_output, OutputFormat


@click.group()
def run():
    """manage run resources."""


@run.command()
@click.option(
    '-e', '--experiment-id', help='Parent experiment ID of listed runs.')
@click.option(
    '--page-token', default='', help="Token for starting of the page.")
@click.option(
    '-m', '--max-size', default=100, help="Max size of the listed runs.")
@click.option(
    '--sort-by',
    default="created_at desc",
    help="Can be '[field_name]', '[field_name] desc'. For example, 'name desc'."
)
@click.option(
    '--filter',
    help=(
        "filter: A url-encoded, JSON-serialized Filter protocol buffer "
        "(see [filter.proto](https://github.com/kubeflow/pipelines/blob/master/backend/api/filter.proto))."
    ))
@click.pass_context
def list(ctx: click.Context, experiment_id: str, page_token: str, max_size: int,
         sort_by: str, filter: str):
    """list recent KFP runs."""
    client = ctx.obj['client']
    output_format = ctx.obj['output']
    response = client.list_runs(
        experiment_id=experiment_id,
        page_token=page_token,
        page_size=max_size,
        sort_by=sort_by,
        filter=filter)
    if response and response.runs:
        _print_runs(response.runs, output_format)
    else:
        if output_format == OutputFormat.json.name:
            msg = json.dumps([])
        else:
            msg = 'No runs found.'
        click.echo(msg)


@run.command()
@click.option(
    '-e',
    '--experiment-name',
    required=True,
    help='Experiment name of the run.')
@click.option('-r', '--run-name', help='Name of the run.')
@click.option(
    '-f',
    '--package-file',
    type=click.Path(exists=True, dir_okay=False),
    help='Path of the pipeline package file.')
@click.option('-p', '--pipeline-id', help='ID of the pipeline template.')
@click.option('-n', '--pipeline-name', help='Name of the pipeline template.')
@click.option(
    '-w',
    '--watch',
    is_flag=True,
    default=False,
    help='Watch the run status until it finishes.')
@click.option('-v', '--version', help='ID of the pipeline version.')
@click.option(
    '-t',
    '--timeout',
    default=0,
    help='Wait for a run to complete until timeout in seconds.',
    type=int)
@click.argument('args', nargs=-1)
@click.pass_context
def submit(ctx: click.Context, experiment_name: str, run_name: str,
           package_file: str, pipeline_id: str, pipeline_name: str, watch: bool,
           timeout: int, version: str, args: List[str]):
    """submit a KFP run."""
    client = ctx.obj['client']
    namespace = ctx.obj['namespace']
    output_format = ctx.obj['output']
    if not run_name:
        run_name = experiment_name

    if not pipeline_id and pipeline_name:
        pipeline_id = client.get_pipeline_id(name=pipeline_name)

    if not package_file and not pipeline_id and not version:
        click.echo(
            'You must provide one of [package_file, pipeline_id, version].',
            err=True)
        sys.exit(1)

    arg_dict = dict(arg.split('=', maxsplit=1) for arg in args)

    experiment = client.create_experiment(experiment_name)
    run = client.run_pipeline(
        experiment.id,
        run_name,
        package_file,
        arg_dict,
        pipeline_id,
        version_id=version)
    if timeout > 0:
        _wait_for_run_completion(client, run.id, timeout, output_format)
    else:
        _display_run(client, namespace, run.id, watch, output_format)


@run.command()
@click.option(
    '-w',
    '--watch',
    is_flag=True,
    default=False,
    help='Watch the run status until it finishes.')
@click.option(
    '-d',
    '--detail',
    is_flag=True,
    default=False,
    help='Get detailed information of the run in json format.')
@click.argument('run-id')
@click.pass_context
def get(ctx: click.Context, watch: bool, detail: bool, run_id: str):
    """display the details of a KFP run."""
    client = ctx.obj['client']
    namespace = ctx.obj['namespace']
    output_format = ctx.obj['output']

    _display_run(client, namespace, run_id, watch, output_format, detail)


def _display_run(client: click.Context,
                 namespace: str,
                 run_id: str,
                 watch: bool,
                 output_format: OutputFormat,
                 detail: bool = False):
    run = client.get_run(run_id).run

    if detail:
        data = {
            key:
            value.isoformat() if isinstance(value, datetime.datetime) else value
            for key, value in run.to_dict().items()
            if key not in ['pipeline_spec'
                          ]  # useless but too much detailed field
        }
        click.echo(data)
        return

    _print_runs([run], output_format)
    if not watch:
        return
    argo_path = shutil.which('argo')
    if not argo_path:
        raise RuntimeError(
            "argo isn't found in $PATH. It's necessary for watch. "
            "Please make sure it's installed and available. "
            "Installation instructions be found here - "
            "https://github.com/argoproj/argo-workflows/releases")

    argo_workflow_name = None
    while True:
        time.sleep(1)
        run_detail = client.get_run(run_id)
        run = run_detail.run
        if run_detail.pipeline_runtime and run_detail.pipeline_runtime.workflow_manifest:
            manifest = json.loads(run_detail.pipeline_runtime.workflow_manifest)
            if manifest['metadata'] and manifest['metadata']['name']:
                argo_workflow_name = manifest['metadata']['name']
                break
        if run_detail.run.status in ['Succeeded', 'Skipped', 'Failed', 'Error']:
            click.echo('Run is finished with status {}.'.format(
                run_detail.run.status))
            return
    if argo_workflow_name:
        subprocess.run(
            [argo_path, 'watch', argo_workflow_name, '-n', namespace])
        _print_runs([run], output_format)


def _wait_for_run_completion(client: Client, run_id: str, timeout: int,
                             output_format: OutputFormat):
    run_detail = client.wait_for_run_completion(run_id, timeout)
    _print_runs([run_detail.run], output_format)


def _print_runs(runs: List[kfp_server_api.ApiRun], output_format: OutputFormat):
    headers = ['run id', 'name', 'status', 'created at', 'experiment id']
    data = [[
        run.id, run.name, run.status,
        run.created_at.isoformat(),
        next(rr
             for rr in run.resource_references
             if rr.key.type == kfp_server_api.ApiResourceType.EXPERIMENT).key.id
    ]
            for run in runs]
    print_output(data, headers, output_format, table_format='grid')



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import itertools
import os
import re
import subprocess
import tempfile
from typing import List
import unittest
from unittest import mock

from absl.testing import parameterized
import click
from click import testing
from kfp.cli import cli
from kfp.cli import compile_


class TestCliNounAliases(unittest.TestCase):

    def setUp(self):
        runner = testing.CliRunner()
        self.invoke = functools.partial(
            runner.invoke, cli=cli.cli, catch_exceptions=False, obj={})

    def test_aliases_singular(self):
        result = self.invoke(args=['component'])
        self.assertEqual(result.exit_code, 0)

    def test_aliases_plural(self):
        result = self.invoke(args=['components'])
        self.assertEqual(result.exit_code, 0)

    def test_aliases_fails(self):
        result = self.invoke(args=['componentss'])
        self.assertEqual(result.exit_code, 2)
        self.assertEqual("Error: Unrecognized command 'componentss'\n",
                         result.output)


class TestAliases(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        runner = testing.CliRunner()
        cls.invoke = functools.partial(
            runner.invoke, cli=cli.cli, catch_exceptions=True, obj={})

    def test_aliases_singular(self):
        result = self.invoke(args=['component'])
        self.assertEqual(result.exit_code, 0)

    def test_aliases_plural(self):
        result = self.invoke(args=['components'])
        self.assertEqual(result.exit_code, 0)

    def test_aliases_fails(self):
        result = self.invoke(args=['componentss'])
        self.assertEqual(result.exit_code, 2)
        self.assertEqual("Error: Unrecognized command 'componentss'\n",
                         result.output)


class TestCliAutocomplete(parameterized.TestCase):

    def setUp(self):
        runner = testing.CliRunner()
        self.invoke = functools.partial(
            runner.invoke, cli=cli.cli, catch_exceptions=False, obj={})

    @parameterized.parameters(['bash', 'zsh', 'fish'])
    def test_show_autocomplete(self, shell):
        result = self.invoke(args=['--show-completion', shell])
        expected = cli._create_completion(shell)
        self.assertIn(expected, result.output)
        self.assertEqual(result.exit_code, 0)

    @parameterized.parameters(['bash', 'zsh', 'fish'])
    def test_install_autocomplete_with_empty_file(self, shell):
        with tempfile.TemporaryDirectory() as tempdir:
            with mock.patch('os.path.expanduser', return_value=tempdir):
                temp_path = os.path.join(tempdir, *cli.SHELL_FILES[shell])
                os.makedirs(os.path.dirname(temp_path), exist_ok=True)

                result = self.invoke(args=['--install-completion', shell])
                expected = cli._create_completion(shell)

                with open(temp_path) as f:
                    last_line = f.readlines()[-1]
                self.assertEqual(expected + '\n', last_line)
                self.assertEqual(result.exit_code, 0)

    @parameterized.parameters(
        list(itertools.product(['bash', 'zsh', 'fish'], [True, False])))
    def test_install_autocomplete_with_unempty_file(self, shell,
                                                    has_trailing_newline):
        with tempfile.TemporaryDirectory() as tempdir:
            with mock.patch('os.path.expanduser', return_value=tempdir):
                temp_path = os.path.join(tempdir, *cli.SHELL_FILES[shell])
                os.makedirs(os.path.dirname(temp_path), exist_ok=True)

                existing_file_contents = [
                    'something\n',
                    'something else' + ('\n' if has_trailing_newline else ''),
                ]
                with open(temp_path, 'w') as f:
                    f.writelines(existing_file_contents)

                result = self.invoke(args=['--install-completion', shell])
                expected = cli._create_completion(shell)

                with open(temp_path) as f:
                    last_line = f.readlines()[-1]
                self.assertEqual(expected + '\n', last_line)
                self.assertEqual(result.exit_code, 0)


class TestCliVersion(unittest.TestCase):

    def setUp(self):
        runner = testing.CliRunner()
        self.invoke = functools.partial(
            runner.invoke, cli=cli.cli, catch_exceptions=False, obj={})

    def test_version(self):
        result = self.invoke(args=['--version'])
        self.assertEqual(result.exit_code, 0)
        matches = re.match(r'^kfp \d\.\d\.\d.*', result.output)
        self.assertTrue(matches)


class TestDslCompile(parameterized.TestCase):

    def invoke(self, args: List[str]) -> testing.Result:
        starting_args = ['dsl', 'compile']
        args = starting_args + args
        runner = testing.CliRunner()
        return runner.invoke(
            cli=cli.cli, args=args, catch_exceptions=False, obj={})

    def invoke_deprecated(self, args: List[str]) -> testing.Result:
        runner = testing.CliRunner()
        return runner.invoke(
            cli=compile_.compile_, args=args, catch_exceptions=False, obj={})

    def test_deprecated_command_is_found(self):
        result = self.invoke_deprecated(['--help'])
        self.assertEqual(result.exit_code, 0)

    def test_deprecation_warning(self):
        res = subprocess.run(['dsl-compile', '--help'], capture_output=True)
        self.assertIn('Deprecated. Please use `kfp dsl compile` instead.)',
                      res.stdout.decode('utf-8'))


info_dict = cli.cli.to_info_dict(ctx=click.Context(cli.cli))
commands_dict = {
    command: list(body.get('commands', {}).keys())
    for command, body in info_dict['commands'].items()
}
noun_verb_list = [
    (noun, verb) for noun, verbs in commands_dict.items() for verb in verbs
]


class TestSmokeTestAllCommandsWithHelp(parameterized.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.runner = testing.CliRunner()

        cls.vals = [('run', 'list')]

    @parameterized.parameters(*noun_verb_list)
    def test(self, noun: str, verb: str):
        with mock.patch('kfp.cli.cli.client.Client'):
            result = self.runner.invoke(
                args=[noun, verb, '--help'],
                cli=cli.cli,
                catch_exceptions=False,
                obj={})
            self.assertTrue(result.output.startswith('Usage: '))
            self.assertEqual(result.exit_code, 0)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Deprecated. See kfp.types.artifact_types instead.

This module will be removed in KFP v2.0.
"""
import warnings
from kfp.components.types import artifact_types

warnings.warn(
    'Module kfp.dsl.io_types is deprecated and will be removed'
    ' in KFP v2.0. Please import types from kfp.dsl instead.',
    category=FutureWarning)

Artifact = artifact_types.Artifact
Dataset = artifact_types.Dataset
Metrics = artifact_types.Metrics
ClassificationMetrics = artifact_types.ClassificationMetrics
Model = artifact_types.Model
SlicedClassificationMetrics = artifact_types.SlicedClassificationMetrics
HTML = artifact_types.HTML
Markdown = artifact_types.Markdown
create_runtime_artifact = artifact_types.create_runtime_artifact



# Copyright 2022 The Kubeflow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import click
from click import testing
from kfp.cli.utils import deprecated_alias_group


@click.group(
    cls=deprecated_alias_group.deprecated_alias_group_factory(
        {'deprecated': 'new'}))
def cli():
    pass


@cli.command()
def new():
    click.echo('Called new command.')


class TestAliasedPluralsGroup(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.runner = testing.CliRunner()

    def test_new_call(self):
        result = self.runner.invoke(cli, ['new'])
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(result.output, 'Called new command.\n')

    def test_deprecated_call(self):
        result = self.runner.invoke(cli, ['deprecated'])
        self.assertEqual(result.exit_code, 0)
        self.assertTrue('Called new command.\n' in result.output)
        self.assertTrue(
            "Warning: 'deprecated' is deprecated, use 'new' instead." in
            result.output)


if __name__ == '__main__':
    unittest.main()
